/*
 * This file contains a copy of the master content sx1276PingPong
 * with adaption for the SX1276Generic environment
 * (c) 2017 Helmut Tschemernjak
 * 30826 Garbsen (Hannover) Germany
 */

#include "mbed.h"
#include "PinMap.h"
#include "GenericPingPong.h"
#include "sx1276-mbed-hal.h"
#include "main.h"
#include <string>

#ifdef FEATURE_LORA

/* Set this flag to '1' to display debug messages on the console */
#define DEBUG_MESSAGE   1

/* Set this flag to '1' to use the LoRa modulation or to '0' to use FSK modulation */
#define USE_MODEM_LORA  1
#define USE_MODEM_FSK   !USE_MODEM_LORA
#define RF_FREQUENCY            RF_FREQUENCY_915_0  // Hz
#define TX_OUTPUT_POWER         20                  // 20 dBm

#if USE_MODEM_LORA == 1

#define LORA_BANDWIDTH          125000  // LoRa default, details in SX1276::BandwidthMap
#define LORA_SPREADING_FACTOR   LORA_SF7
#define LORA_CODINGRATE         LORA_ERROR_CODING_RATE_4_5

#define LORA_PREAMBLE_LENGTH    8       // Same for Tx and Rx
#define LORA_SYMBOL_TIMEOUT     5       // Symbols
#define LORA_FIX_LENGTH_PAYLOAD_ON  false
#define LORA_FHSS_ENABLED       true
#define LORA_NB_SYMB_HOP        4
#define LORA_IQ_INVERSION_ON    false
#define LORA_CRC_ENABLED        true

#elif USE_MODEM_FSK == 1

#define FSK_FDEV                25000     // Hz
#define FSK_DATARATE            19200     // bps
#define FSK_BANDWIDTH           50000     // Hz
#define FSK_AFC_BANDWIDTH       83333     // Hz
#define FSK_PREAMBLE_LENGTH     5         // Same for Tx and Rx
#define FSK_FIX_LENGTH_PAYLOAD_ON   false
#define FSK_CRC_ENABLED         true

#else
#error "Please define a modem in the compiler options."
#endif


#define RX_TIMEOUT_VALUE    3500	// in ms

//#define BUFFER_SIZE       32        // Define the payload size here
#define BUFFER_SIZE         256        // Define the payload size here

/*
 *  Global variables declarations
 */
typedef enum {
    LOWPOWER = 0,
    IDLE,

    RX,
    RX_TIMEOUT,
    RX_ERROR,

    TX,
    TX_TIMEOUT,

    CAD,
    CAD_DONE
} AppStates_t;

volatile AppStates_t State = LOWPOWER;

/*!
 * Radio events function pointer
 */
static RadioEvents_t RadioEvents;

/*
 *  Global variables declarations
 */
SX1276Generic *Radio;


//const uint8_t PingMsg[] = { 0xff, 0xff, 0x00, 0x00, 'P', 'I', 'N', 'G'};// "PING";
//const uint8_t PongMsg[] = { 0xff, 0xff, 0x00, 0x00, 'P', 'O', 'N', 'G'};// "PONG";
//const uint8_t PingMsg[] = { 'M', 'S', 'J', ' ','G','I','L'};// "PING";
//const uint8_t PongMsg[] = {  'R', 'P', 'T', 'A',' ','G','I','L'};// "PONG";

//construccion de la trama
const char EUI[] = "0A01";// 4  bytes que definen el identificador del dispositivo (copiar los bytes en string)
const char AppEUI[] = "AAAA";// 4  bytes que definen el identificador de la aplicacion (copiar los bytes en string)
const char AppKey[] = "1A1B";// 4  bytes que definen la clave de la aplicacion (copiar los bytes en string) - El protocolo LoRaWAN establece la clave de 16 bytes pero para efectos de prueba se hara de 4
char MsgTX[256] = "";// Mensaje de transmision, se pueden usar los 52 bytes faltantes para completar el payload de 64 bytes. Se puede poner directamente en string.
char MsgRX[256] = "";// Mensaje de recepcion, carga el payload entrante a esta cadena.
char MsgRet[] = "RECIBIDO";
char DestEUI[4] = "";
char euiAux[4]="";
uint16_t BufferSize = BUFFER_SIZE;
uint8_t *Buffer;
string msjDeco="";
char *retParse;
char *srcEUI;
char *msjDestEUI;
char *msjContent;
Serial serRPi(PA_9, PA_10);

DigitalOut *led3;


int SX1276PingPong()
{
#if( defined ( TARGET_KL25Z ) || defined ( TARGET_LPC11U6X ) )
    DigitalOut *led = new DigitalOut(LED2);
#elif defined(TARGET_NUCLEO_L073RZ) || defined(TARGET_DISCO_L072CZ_LRWAN1)
    DigitalOut *led = new DigitalOut(LED4);   // RX red
    led3 = new DigitalOut(LED3);  // TX blue
#else
    DigitalOut *led = new DigitalOut(LED1);
    led3 = led;
#endif

    Buffer = new  uint8_t[BUFFER_SIZE];
    *led3 = 1;

#ifdef B_L072Z_LRWAN1_LORA
    Radio = new SX1276Generic(NULL, MURATA_SX1276,
                              LORA_SPI_MOSI, LORA_SPI_MISO, LORA_SPI_SCLK, LORA_CS, LORA_RESET,
                              LORA_DIO0, LORA_DIO1, LORA_DIO2, LORA_DIO3, LORA_DIO4, LORA_DIO5,
                              LORA_ANT_RX, LORA_ANT_TX, LORA_ANT_BOOST, LORA_TCXO);
#else // RFM95
    Radio = new SX1276Generic(NULL, RFM95_SX1276,
                              LORA_SPI_MOSI, LORA_SPI_MISO, LORA_SPI_SCLK, LORA_CS, LORA_RESET,
                              LORA_DIO0, LORA_DIO1, LORA_DIO2, LORA_DIO3, LORA_DIO4, LORA_DIO5);

#endif
	//serRPi = new BufferedSerial(PA_9, PA_10);
    serRPi.baud(115200);
    serRPi.format(8);
    uint8_t i;
    bool isServer = true;

    dprintf("Aplicacion de comunicacion LoRa punto a punto" );
    dprintf("Frecuencia: %.1f", (double)RF_FREQUENCY/1000000.0);
    dprintf("TXPower: %d dBm",  TX_OUTPUT_POWER);
#if USE_MODEM_LORA == 1
    dprintf("Bandwidth: %d Hz", LORA_BANDWIDTH);
    dprintf("Spreading factor: SF%d", LORA_SPREADING_FACTOR);
#elif USE_MODEM_FSK == 1
    dprintf("Bandwidth: %d kHz",  FSK_BANDWIDTH);
    dprintf("Baudrate: %d", FSK_DATARATE);
#endif
    // Initialize Radio driver
    RadioEvents.TxDone = OnTxDone;
    RadioEvents.RxDone = OnRxDone;
    RadioEvents.RxError = OnRxError;
    RadioEvents.TxTimeout = OnTxTimeout;
    RadioEvents.RxTimeout = OnRxTimeout;
    if (Radio->Init( &RadioEvents ) == false) {
        while(1) {
            dprintf("Radio could not be detected!");
            wait( 1 );
        }
    }


    switch(Radio->DetectBoardType()) {
        case SX1276MB1LAS:
            if (DEBUG_MESSAGE)
                dprintf(" > Board Type: SX1276MB1LAS <");
            break;
        case SX1276MB1MAS:
            if (DEBUG_MESSAGE)
                dprintf(" > Board Type: SX1276MB1LAS <");
        case MURATA_SX1276:
            if (DEBUG_MESSAGE)
                dprintf(" > Board Type: MURATA_SX1276_STM32L0 <");
            break;
        case RFM95_SX1276:
            if (DEBUG_MESSAGE)
                dprintf(" > HopeRF RFM95xx <");
            break;
        default:
            dprintf(" > Board Type: unknown <");
    }

    Radio->SetChannel(RF_FREQUENCY );

#if USE_MODEM_LORA == 1

    if (LORA_FHSS_ENABLED)
        dprintf("             > LORA FHSS Mode <");
    if (!LORA_FHSS_ENABLED)
        dprintf("             > LORA Mode <");

    Radio->SetTxConfig( MODEM_LORA, TX_OUTPUT_POWER, 0, LORA_BANDWIDTH,
                        LORA_SPREADING_FACTOR, LORA_CODINGRATE,
                        LORA_PREAMBLE_LENGTH, LORA_FIX_LENGTH_PAYLOAD_ON,
                        LORA_CRC_ENABLED, LORA_FHSS_ENABLED, LORA_NB_SYMB_HOP,
                        LORA_IQ_INVERSION_ON, 2000 );

    Radio->SetRxConfig( MODEM_LORA, LORA_BANDWIDTH, LORA_SPREADING_FACTOR,
                        LORA_CODINGRATE, 0, LORA_PREAMBLE_LENGTH,
                        LORA_SYMBOL_TIMEOUT, LORA_FIX_LENGTH_PAYLOAD_ON, 0,
                        LORA_CRC_ENABLED, LORA_FHSS_ENABLED, LORA_NB_SYMB_HOP,
                        LORA_IQ_INVERSION_ON, true );

#elif USE_MODEM_FSK == 1

    dprintf("              > FSK Mode <");
    Radio->SetTxConfig( MODEM_FSK, TX_OUTPUT_POWER, FSK_FDEV, 0,
                        FSK_DATARATE, 0,
                        FSK_PREAMBLE_LENGTH, FSK_FIX_LENGTH_PAYLOAD_ON,
                        FSK_CRC_ENABLED, 0, 0, 0, 2000 );

    Radio->SetRxConfig( MODEM_FSK, FSK_BANDWIDTH, FSK_DATARATE,
                        0, FSK_AFC_BANDWIDTH, FSK_PREAMBLE_LENGTH,
                        0, FSK_FIX_LENGTH_PAYLOAD_ON, 0, FSK_CRC_ENABLED,
                        0, 0, false, true );

#else

#error "Please define a modem in the compiler options."

#endif

    if (DEBUG_MESSAGE)
        dprintf("Inicializando el servidor");
    dprintf("EUI (ID): %s",EUI);

    Radio->Rx( RX_TIMEOUT_VALUE );

    while( 1 ) {
#ifdef TARGET_STM32L4
        WatchDogUpdate();
#endif

        switch( State ) {
            case RX:
                //dprintf( "Recibiendo informacion" );
                *led = !*led;
                dprintf("Mensaje para depurar: %s",MsgRX);
                //strcpy(DestEUI,"A010");
                msjDeco=MsgRX;
                //vector<string> x = split(msjDeco, '|');
                splitOnPosition(MsgRX, 0);
                //msjDestEUI=splitOnPosition(MsgRX, 3);
                //msjContent=splitOnPosition(MsgRX, 4);

                dprintf("Source EUI: %s, Destination EUI: %s, Content: %s",srcEUI,msjDestEUI,msjContent);
                if(strcmp(EUI,msjDestEUI) == 0) {
                    dprintf("Mismo EUI, Soy el destinatario");

                } else {
                    dprintf("Diferente EUI, ignorar mensaje");
                    Radio->Rx( RX_TIMEOUT_VALUE );
                    wait_ms( 500 );
                    *led = !*led;
                    State = LOWPOWER;
                    break;
                }
                strcpy(MsgRet,"RECIBIDO");
                strcpy(MsgTX, EUI);
                strcat(MsgTX, "|");
                strcat(MsgTX, AppEUI);
                strcat(MsgTX, "|");
                strcat(MsgTX, AppKey);
                strcat(MsgTX, "|");
                strcat(MsgTX, srcEUI);
                strcat(MsgTX, "|");
                strcat(MsgTX, MsgRet);
                //strcat(MsgTX, ";");
                //MsgRX = "RECIBIDO";
                memcpy(Buffer, MsgTX, sizeof(MsgTX));
                // Se llena el buffer con la informacion a enviar
                for( i = sizeof(MsgTX); i < BufferSize; i++ ) {
                    Buffer[i] = i - sizeof(MsgTX);
                }
                //dprintf("enviando mensaje a raspberry");
                serRPi.printf("contenido: %s \r\n",msjContent);
                Radio->Send( Buffer, BufferSize );
                Radio->Rx( RX_TIMEOUT_VALUE );
                wait_ms( 500 );
                
                /*Enviar contenido a la raspberry por puerto serial*/
                
                
                /***************************************************/
                
                *led = !*led;
                State = LOWPOWER;
                break;
            case TX:
                *led3 = !*led3;
                if( isServer == true ) {
                    //dprintf("Enviando respuesta al nodo" );
                    //dprintf("Mensaje a enviar: %s",MsgTX);
                }
                wait_ms( 500 );
                *led3 = !*led3;
                Radio->Rx( RX_TIMEOUT_VALUE );
                State = LOWPOWER;
                break;
            case RX_TIMEOUT:

                Radio->Rx( RX_TIMEOUT_VALUE );
                State = LOWPOWER;
                break;
            case RX_ERROR:
                // We have received a Packet with a CRC error, send reply as if packet was correct
                /*strcpy(MsgRet,"ERROR");
                strcpy(MsgTX, EUI);
                strcat(MsgTX, "|");
                strcat(MsgTX, AppEUI);
                strcat(MsgTX, "|");
                strcat(MsgTX, AppKey);
                strcat(MsgTX, "|");
                strcat(MsgTX, "A010");
                strcat(MsgTX, "|");
                strcat(MsgTX, MsgRet);
                //MsgRX = "ERROR";
                if( isServer == true ) {
                    // Se llena el buffer para el envio del error
                    memcpy(Buffer, MsgTX, sizeof(MsgTX));
                    for( i = sizeof(MsgTX); i < BufferSize; i++ ) {
                        Buffer[i] = i - sizeof(MsgTX);
                    }
                    wait_ms( 1000 );
                    Radio->Send( Buffer, BufferSize );
                }*/
                Radio->Rx( RX_TIMEOUT_VALUE );
                State = LOWPOWER;
                break;
            case TX_TIMEOUT:
                Radio->Rx( RX_TIMEOUT_VALUE );
                State = LOWPOWER;
                break;
            case LOWPOWER:
                sleep();
                break;
            default:
                State = LOWPOWER;
                break;
        }
    }
}

char *splitOnPosition(char *msj, int pos)
{

    int i=0;
    char *substring = strtok (msj,"|");
    char *strOutput="";
    while (substring != NULL) {
        //dprintf ("substring: %s, index: %d",substring,i);

        if(i == 0) {
            srcEUI = substring;
        } else if(i == 3) {
            msjDestEUI = substring;
        } else if(i == 4) {
            msjContent = substring;
        } else if(i > 4) {
            strcat(msjContent," ");
            strcat(msjContent,substring);
            //dprintf ("contenido: %s, substring: %s, i: %d",msjContent, substring, i);
        }
        substring = strtok (NULL, "|");
        i++;
    }
    return strOutput;
}

void OnTxDone(void *radio, void *userThisPtr, void *userData)
{
    Radio->Sleep( );
    State = TX;
    if (DEBUG_MESSAGE)
        dprintf("> OnTxDone");
}

void OnRxDone(void *radio, void *userThisPtr, void *userData, uint8_t *payload, uint16_t size, int16_t rssi, int8_t snr)
{
    Radio->Sleep( );
    BufferSize = size;
    memcpy( Buffer, payload, BufferSize );
    State = RX;
    if (DEBUG_MESSAGE)
        dprintf("> OnRxDone: RssiValue=%d dBm, SnrValue=%d", rssi, snr);
    //dump("Data:", payload, size);
    strcpy(MsgRX,(char*)payload);
    //dprintf("Msj: %s", MsgRX);
}

void OnTxTimeout(void *radio, void *userThisPtr, void *userData)
{
    *led3 = 0;
    Radio->Sleep( );
    State = TX_TIMEOUT;
    if(DEBUG_MESSAGE)
        dprintf("> OnTxTimeout");
}

void OnRxTimeout(void *radio, void *userThisPtr, void *userData)
{
    *led3 = 0;
    Radio->Sleep( );
    Buffer[BufferSize-1] = 0;
    State = RX_TIMEOUT;
    if (DEBUG_MESSAGE)
        dprintf("> OnRxTimeout");
}

void OnRxError(void *radio, void *userThisPtr, void *userData)
{
    Radio->Sleep( );
    State = RX_ERROR;
    if (DEBUG_MESSAGE)
        dprintf("> OnRxError");
}

#endif
